

let soi = string_of_int
(* Les types *)

type expr =
    Zero
  |   App of expr * bag
  |   Var of string
  |   Cons of string
  |   Abs of string * expr
and bag = (expr*int) list


type closure=C of bag*env
and env= E of int*int
and stack= closure list


type instr =
     Grab of string          (* met le sommet de la pile dans l'environnement *)
|    Push of bag   (*  *)
|    Access of string  (* acces \`a la n-ieme variable de l'environnement *)
|    Go_under of string


(*  Les constantes  *)

let empty = E(0,-1) 
let infinity = -27
let max_loop = 20
let largeur_max =100
let taille_max = 100
let nb_env = ref 0


let genv = Array.make_matrix largeur_max taille_max  ("$",C([],E(-1,-1)),E(-1,-1)) (* type (string * closure * env) array *)
let pointeurs = Array.make_matrix largeur_max taille_max 0
let count_env=Array.make largeur_max 0

let next_free () = 
    let k=ref 0 in
    while count_env.(!k)>0 do
        incr k  
    done;
    !k

(* Les fonctions d'affichage  *)


let rec print_expr exp =
  let rec print_rec = function
    |Zero -> print_string "** 0 **"
    |Var(s)-> print_string (s)
    |Cons(s) -> print_string ("$"^s)
    |App(u,v)-> print_string "<"; print_rec u; print_string ">"; print_bag v; print_string ""
    |Abs(x,f) -> (match f with
	 	    | Abs(y,s)-> print_rec (Abs(x^y,s))
		    | _ ->  print_string ("λ"^x^"."); print_rec f  )
  in print_rec exp
and print_bag bag =
  let bag_aux = function
    |a,n when n=infinity -> print_string "!"; print_expr a
    |a,n ->  print_expr a; print_string ("*"^string_of_int(n))
  in
  let rec aux =function
    | []->()
    | [a] -> bag_aux a
    | a::b -> bag_aux a; print_string ", "; aux b
  in
    print_string "["; aux bag; print_string "]"

let rec print_somme = function
  |[] -> ()
  |[a] -> print_expr a
  |Zero::b->print_somme b
  |a::b -> print_expr a; print_string " + "; print_somme b

let rec print_env = function
  | E(i,-1) -> print_string "∅"
  | E(i,n)-> let (s,C(bag,E(xc,yc)),E(xn,yn))= (genv.(i).(n)) in   
      print_string ("E("^string_of_int(i)^","^string_of_int(n)^"):["^s^"->("); print_bag bag; print_string ","; print_env (E(xc,yc)); print_string ")]·";print_env (E(xn,yn))


let rec print_env_final = function
  | E(i,-1) -> print_string "∅"
  | E(i,n)-> let (s,C(bag,E(xc,yc)),E(xn,yn))= (genv.(i).(n)) in   
      print_string ("["^s^"->("); print_bag bag; print_string ","; print_env_final (E(i,yc)); print_string ")]·";print_env_final (E(i,yn))

let print_env_total n = 
    for i=0 to count_env.(n) do
        let (s,C(bag,E(xc,yc)),E(xn,yn))= (genv.(n).(i)) in   
        print_string ("E("^string_of_int(i)^"):["^s^"->(");
        print_bag bag; print_string ",";
        print_string ("E("^soi(yc)^")");
        print_string ")] -- "
    done

let rec print_stack stack (E(n,k))= match stack with
  | []->print_string "∅"
  | C(b,E(xc,yc))::t -> print_string "C("; print_bag b; print_string "," ;
  print_env (E(n,yc)); print_string ")::"; print_stack t (E(n,k))


let print_instr = function
  | Grab(s) -> print_string ("Grab("^s^")  ")
  | Push(b) -> print_string "Push("; print_bag b; print_string ")  "
  | Access("fin")-> print_string ("***\tResult:\t"); failwith "End"
  | Access("begin") -> print_string "Beginning"
  | Access(s) ->  print_string ("Access("^s^")")
  | Go_under(s) ->  print_string ("Go_under("^s^")")



let print_tab tab=
  for i=1 to tab do
    print_string ">\t"
  done

  
let print_ref () =
    for i = 0 to !nb_env do
        for j=0 to count_env.(i) do
            print_string " * ";
            print_int (pointeurs.(i).(j));
        done;
        print_newline()
    done
	  
let print_head head=
  let rec aux = function
  |[] -> print_string " ~ "
  |a::b -> print_string ("λ"^a^"."); aux b
  in aux (List.rev head)

let rec print_compile tab head= function (instr,expr,stack,env)->
  try 
    (print_tab tab;
     print_instr instr;
     print_string "\t|\t";
     print_head head;
     print_expr expr;
     print_string "\t|\t";
     print_stack stack env;
     print_string "\t|\t";
     print_env env; 
     print_newline()
(*     print_ref ()*)
    )
  with _ -> print_expr expr;print_string "\t***\n" 


(*  Les primitives spécifiques aux environnements *)


let rec union (E(a1,e1)) (E(a2,e2))=
      match genv.(a1).(e1) with   (* /!\ bug si e1 = -1  /!\ *)
      | s,clos,E(a,-1)-> genv.(a1).(e1)<-s,clos,E(a2,e2)
      | s,clos,E(m,n) -> union (E(m,n)) (E(a2,e2))


let clean () = 
  for k=0 to (Array.length genv)-1 do
    for j=0 to (Array.length genv.(k))-1 do
        genv.(k).(j)<-("$",C([],E(-1,-1)),E(-1,-1));
      pointeurs.(k).(j)<-0
    done
  done;
  for k=0 to (Array.length count_env)-1 do
    count_env.(k)<-0
  done


let rec len_env = function
  | E(i,-1) -> 0
  | E(i,n) -> let _,_,e = genv.(i).(n) in 1 + len_env e

let copy_env i j = (* duplique E(i,_) dans E(j,_) *)
  count_env.(j)<- count_env.(i);
  for k=0 to count_env.(j) do
    pointeurs.(j).(k)<-pointeurs.(i).(k); (* dédouble les pointeurs *)
    let a,C(b,E(x1,y1)),E(x2,y2) = genv.(i).(k) in
      genv.(j).(k) <- a,C(b,E(j,y1)),E(j,y2)
  done


(* Le fourre-tout  *)

let minic = function 
  | m,n when m = infinity -> n
  | m,n when n = infinity -> m
  | m,n -> min m n

let mini m n = minic (m,n );;


exception Empty_bag

let head = function
  |[] -> raise Empty_bag
  |a::b -> a

let tail = function
  |[] -> raise Empty_bag
  |a::b -> b

(* Analyse statique *)


let occurences terme=
  let rec count var= function
    |Zero -> 0
    |Var x when x=var -> 1
    |Var x -> 0
    |Cons x -> 0
    |App(u,t)->(let rec aux =function
		  |[]-> count var u
		  |(a,n)::b when n=infinity-> failwith "infinity"
		  |(a,n)::b -> n*(count var a) + (aux b)
		in aux t)
    |Abs(x,t) when x=var-> 0
    |Abs(_,t)->count var t
  in
    match terme with
      |Zero -> 0
      |Var _ -> -1
      |Cons _ -> -1	    
      |App(_,_) -> -1
      |Abs(var,t)->try count var t with _ -> infinity
 
    
let rec count_linear=function
  |[] -> 0
  |(a,n)::q when n=infinity-> count_linear q
  |(a,n)::q -> n + (count_linear q);;

(* Pour le comptage de références ! *)
    
let rec parcours_env f= function
  |E(_,-1)-> ();
  |E(k,n) -> let x,C(bag,envb),next_env=genv.(k).(n) in
      f k n ; (* action au choix *)
      parcours_env f envb;
      parcours_env f next_env

let rec parcours_stack f = function
  |[]-> ();
  |C(bag,envc)::t -> parcours_env f envc;
      parcours_stack f t

let add_ref k n =
  pointeurs.(k).(n)<-pointeurs.(k).(n) + 1

let soi=string_of_int

exception Overfed of string

let sub_ref k n =
  pointeurs.(k).(n)<-pointeurs.(k).(n) - 1;
  if pointeurs.(k).(n)=0 
  then (
    let (s,C(bag,_),_)=genv.(k).(n) in
      if count_linear bag > 0 
      then raise (Overfed s)
  )
